package dcsp;
import dcsp.parameters.*;
import java.io.*;
import java.rmi.*;

class Agent {
    String name;
    String fileName="";
    String url = "rmi:///BrokerOfTheLIA_SearchMachine";
    public IBrokerServer broker;
    AgentServer skel;
    AgentGUI myGUI;
    Problem problem;
    Object problemMonitor= new Object();
    public Semaphore agentSemaphore=new Semaphore();
    boolean debug;
    int importance=0;
    String[] protocols;

    Agent (String title, String _url){
        name = title;
        protocols= new String[1];
        protocols[0]="_ABT_";
	System.out.println("before alive, wait for problemMonitor");
	synchronized(problemMonitor){
	    problem=new Problem(this);
	}
        System.out.println("agent alive!");
        try {
            url = System.getProperty("broker",url);
        } catch (Exception e){
	    System.err.print("Agent:Agent:getProperty:");
	    System.err.println(e);
	    System.err.println("Error getting properties");
	    e.printStackTrace();
	}
        if (!(_url.equals("none")))
            url=_url;
        System.out.println("Looking for url: " + url);
        try {
            broker = (IBrokerServer) Naming.lookup(url);
            System.out.println("Locating remote object...! " + url);
        }catch (Exception e){
	    System.err.println(e);
	    System.err.println("Agent:Agent:Error while locating broker: " + url);
	    e.printStackTrace();
	}
        try {
            skel = new AgentServer(this);
        }catch (Exception e){
	    System.err.println(e);
	    System.err.println("Agent:Agent:Error launching server");
	    e.printStackTrace();
	}
        //myGUI = new AgentGUI(title, this);
	participate();
    }
    public Object message(String command, Object parameters){
        if(command.equals("LOAD")) {
	    //System.out.println(command+" waits for problemMonitor");
	    synchronized(problemMonitor) {
		//System.out.println(command+" in problemMonitor");
		problemReset();
		try{
		    System.out.println("Gonna remove threads");
		    skel.removeThreads();
		    System.out.println("Remove threads done");
		}catch(Exception e) {
		    System.err.println("Agent:message:rT:");
		    System.err.println(e);
		}
		fileName=(String)parameters;
		if(loadFile(new File(fileName)))
		    run();
		else
		    System.err.println("Not able to load file:"+fileName);
	    }
	    //System.out.println(command+" out of problemMonitor");
	    return null;
	}
        if(command.equals("GETNAME")){
	    return getName();
	}
        if(command.equals("EXIT")){
	    System.out.println("Fin: "+parameters.toString());
	    System.err.println("Fin: "+parameters.toString());
	    System.exit(0);
	}
        if(command.equals("STEP")){
	    if(debug)
		step();
	    return new DOk();
	}
	Object res;
	Problem tmp;
	//System.out.println(command+" waits for problemMonitor");
	synchronized(problemMonitor){
	    //System.out.println(command+" in problemMonitor");
	    if(problem==null){
		System.out.println(command+" meets Empty problem->exit");
		return null;
	    }
	    tmp=problem;
	}
	res=tmp.message(command, parameters);
   
	//System.out.println(command+" out of problemMonitor");
	return res;
    }
    public void step(){
        if(!debug)
            debug=true;
        else
            synchronized (agentSemaphore){
		agentSemaphore.inc();
		agentSemaphore.notify();
	    }       
    }
    public boolean waitStep(){
        if(debug){
            System.out.print("Wait step...");
            if(agentSemaphore.requestWait()){
                System.out.println("restart");
                return true;
            }
            System.out.println("killed");
            return false;
        }
        return true;
    }
    public void problemReset(){
        problem=new Problem(this);
    }
    public IAgentServer getSkel(){
        return skel;
    }
    public void run(){
	if(problem==null){
	    return;
	}
        int complexity=problem.getComplexity();
        try {
            SVariable[] res;
            res=problem.getSVariables();
            DSubscription params= new DSubscription(skel, res, importance, complexity, protocols, getName());
            AgentServer.oneWaySend(broker, "SUBSCRIBE", params);
            System.err.println("Subscribing to broker: "+ url + "...DONE!"+fileName);
	    res=null;
	    params=null;
        }catch (Exception e){
	    System.err.println(e);
	    System.err.println("Agent:run:Error while calling broker");
	    e.printStackTrace();
	}
    }
    void participate(){
        try {
            AgentServer.oneWaySend(broker, "PARTICIPATE", skel);
            System.out.println("Participation to broker: "+ url + "...DONE!");
        }catch (Exception e){
	    System.err.println("Agent:participate:Error while calling broker to participate");
	    System.err.println(e);
	    e.printStackTrace();
	}
    }
    String getName(){
        return name;
    }
    void setName(String _name){
        name=_name;
    }
    boolean loadFile(File file){
        if (!file.exists()) {
	    try{
		System.err.println("loadFile:Inexisting file:"+file.getCanonicalPath());
	    }catch(Exception e){
		System.err.println("loadFile:Inexisting file error");
	    }
            return false;
	}
        if (!problem.loadFile(file)) {
	    System.err.println("loadFile:Error loading file");
            return false;
	}
	fileName=file.getName();
	if(myGUI!=null)
	    myGUI.loaded(file);
        return true;
    }
    IBrokerServer getBroker(){
        return broker;
    }
    public void finalize() throws Throwable{
        System.out.println("agent dying!");
        super.finalize();
    }
}
