package dcsp;
import dcsp.parameters.*;
import java.util.ArrayList;
import java.rmi.*;
import java.rmi.server.*;
import java.net.*;

class AgentServer extends UnicastRemoteObject implements IAgentServer {
    static DInt roundID=new DInt(0);
    static void oneWaySend(IAgentServer target, String command, Object params) throws RemoteException{
	target.oneWayMessage(command,params,(Object)roundID);
    }
    static void oneWaySend(IAgentServer target, String command, Object params, Object sID) throws RemoteException{
	target.oneWayMessage(command,params,sID);
    }
    static void oneWaySend(IBrokerServer target, String command, Object params) throws RemoteException{
	target.oneWayMessage(command,params,(Object)roundID);
    }
    static void oneWaySend(IBrokerServer target, String command, Object params, Object sID) throws RemoteException{
	target.oneWayMessage(command,params,sID);
    }
    static Object messageSend(IAgentServer target, String command, Object params) throws RemoteException{
	return target.message(command,params,(Object)roundID);
    }
    Agent agent;
    // a list of the active OneWayThreads
    ArrayList threads=new ArrayList();
    AgentServer(Agent _agent) throws RemoteException, MalformedURLException{
	super();
	agent= _agent;
	Naming.rebind(agent.getName(), this);
    }
    public String toString(){
	return agent.getName();
    }
    private boolean waitCommand(Object k){
	synchronized(roundID){
	    while(((DInt)k).val>roundID.val){
		try{
		    roundID.wait(10000);
		}catch(Exception e){
		    e.printStackTrace();
		    System.err.println("AgentServer:Interrupted Command");
		    return false;
		}
	    }
	    if(((DInt)k).val<roundID.val)
		return false;
	}
	return true;
    }
    // 1=pre 0=actual -1=old
    private int upToDateCommand(String command, Object k){
	System.out.println("Position incoming:"+((DInt)k).val+" old:"+roundID.val);
	if(command.equals("LOAD")) {
	    synchronized(roundID){
		roundID.val=((DInt)k).val;
		roundID.notifyAll();
	    }
	}
	if(((DInt)k).val<roundID.val){
	    System.out.println(command+" Outdated");
	    return -1;
	}
	if(((DInt)k).val==roundID.val){
	    System.out.println(command+" UpToDate");
	    return 0;
	}
	return 1;
    }
    public Object message(String command, Object parameters,Object sID) throws RemoteException{
	System.out.print("Server: "+command);
	if (parameters!=null)
	    System.out.println(" "+parameters.toString());
	System.out.println("");
	//
	int status=upToDateCommand(command,sID);
	if(status<0)
	    return null;
	if(status>0){
	    if(!waitCommand(sID))
		return null;
	}
	//
	try {
	    return agent.message(command, parameters);
	}
	catch (Exception e) {
	    e.printStackTrace();
	    return null;
	}
    }
    public void oneWayMessage(String command, Object parameters, Object sID) throws RemoteException{
	//
	int status=upToDateCommand(command,sID);
	if(status<0)
	    return;
	if(status>0){
	    if(!waitCommand(sID))
		return;
	}
	//
	OneWayThread omMessage=new OneWayThread(this);
	omMessage.setMessage(command, parameters, new DInt(roundID.val));
	omMessage.removable=true;
	synchronized(threads){
	    threads.add(omMessage);
	    omMessage.start();
	}
	return;
    }
    // to be called at SEARCH STOP
    void removeThreads(){
	OneWayThread th;
	OneWayThread myself=null;
	System.out.println("Wait for threads");
	synchronized(threads){
	    while(!threads.isEmpty()){
		System.out.println("One more zombie");
		th=(OneWayThread)threads.remove(0);
		if(th==OneWayThread.currentThread()){
		    myself=th;
		    continue;
		}
		if(!th.isAlive()){
		    System.err.println("Thread was dead!");
		    continue;
		}
		System.out.println("interrupting thread list");
		th.interrupt();
	    }
	    if (myself!=null)
		threads.add(myself);
	}
	System.out.println("End removing threads");
    }
    void removeThread(OneWayThread th){
	synchronized(threads){
	    int id=threads.lastIndexOf(th);
	    //already killed(thought zombie)
	    if(id<0)
		return;
	    threads.remove(id);
	}
    }
    /*
    public void addLink(IAgentServer toLink) throws RemoteException
    {
	try {
	    SAgent[] agents=new SAgent[1];
	    agents[0]=new SAgent(toLink);
	    registerOutGoing(agents);
	}
	catch (Exception e)
	    {
		e.printStackTrace();
	    }
    } 
       
    public void sendNogood(String to, int context) throws RemoteException
    {
	IAgentServer pier=(IAgentServer)agent.getBroker().message("GETBYNAME", new DString(to));
	pier.oneWayMessage("NOGOOD", new DInt(context));
    }
    */
}
