package dcsp;
import dcsp.parameters.*;
import java.rmi.*;
import java.net.*;
import java.util.Hashtable;
import java.util.*;

class Broker {
    String name;
    boolean debug=false, running=false;

    String protocol;
    //String defaultProtocol="Synchron";
    String protocols[]; // list of all protocols
    String defaultProtocol="C_ABT_";
    ProtocolLeader protocolLeader;
    TestLeader testLeader=new TestLeader(this);
    BrokerGUI myGUI;
    IBrokerServer brokerServer;
    // search hash mapping IAgentServer to AgentData
    Hashtable problemByAgent= new Hashtable();
    // search hash mapping String to ArrayList(SAgent)
    Hashtable problemByVariable= new Hashtable();
    // search hash mapping String names to IAgentServer
    Hashtable problemByAgentName= new Hashtable();
    /*
      IAgentServer[] sortedAgents= new IAgentServer[0];
      // search hash mapping IAgentServer to position
      Hashtable agentOrder= new Hashtable();
    */

    public Semaphore leaderSemaphore=new Semaphore();
    Object protocolLeaderMonitor=new Object();
    Object subscriptionMonitor=new Object();

    Broker(String title){
	//setDaemon();
	name = title;
	protocol=defaultProtocol;
	setProtocol(defaultProtocol);
	System.out.println("broker alive!");

	try{
	    brokerServer = new BrokerServer(this);
	    System.out.println("BrokerServer: "+name+ " created!");
	}
	catch (Exception e){
	    System.err.println(e);
	    System.err.println("Broker:BrokerServer: "+name+ " error");
	    System.err.println(e.getMessage());
	    e.printStackTrace();
	}
	
	try {
	    Naming.rebind(getName(), brokerServer);
	    System.out.println("BrokerServer: "+name+" ready");
	}
	catch (Exception e){
	    System.err.println(e);
	    System.err.println("Binding: "+name+ " error");
	    System.err.println(e.getMessage());
	    e.printStackTrace();
	}
	protocols=new String[4];
	protocols[0]="C_ABT_";
	protocols[1]="_ABT_";
	protocols[2]="I_ABT_";
	protocols[3]="U_ABT_";

	myGUI = new BrokerGUI(title, this);
	if(myGUI==null)
	    System.err.println("No GUI got");
	//System.gc();
	//System.runFinalization();

    }
    String[] getProtocols(){
	return protocols;
    }
    // called at empty nogood or at solution
    void endSolving(String result){
	testLeader.endSolving(result);
    }
    Object message(String command, Object parameters) throws RemoteException{
	if(command.equals("GETBYNAME")){
	    return getByName(((DString)parameters).get());
	}
	if(command.equals("SUBSCRIBE")){
	    if(running){
		System.out.println("Subscription Failure");
		return new DBoolean(false);
	    }
	    boolean result;
	    DSubscription params= (DSubscription) parameters;
	    synchronized(subscriptionMonitor){
		System.out.println("agent "+params.name+ " importance " + params.importance+ " complexity "+params.complexity);
		result=subscribe(params.agent,
				 params.variables,
				 params.importance,
				 params.complexity
				 );
	    }
	    testLeader.subscribe();
	    return new DBoolean(result);
	}
	if (command.equals("PARTICIPATE")){
	    testLeader.add((IAgentServer)parameters);
	    return null;
	}
	if (command.equals("IAMCLEAN")){
	    testLeader.clean();
	    return null;
	}
	if (command.equals("TESTS")){
	    tests_run();
	    return null;
	}
	if (command.equals("PROTOCOL ERROR")){
	    System.err.println("agent "+((DString)parameters).get()+" don't implement protocol "+protocol);
	    clearSession("Agent \""+((DString)parameters).get()+"\"do not implement correctly protocol \""+protocol+"\"");
	}
	ProtocolLeader tmp;
	synchronized(protocolLeaderMonitor){
	    tmp=protocolLeader;
	    if(tmp==null)
		return null;
	}
	return tmp.message(command, parameters);
    }
    void distribute(IAgentServer ag, String fileName){
	try{
	    BrokerServer.oneWaySend(ag,"LOAD",fileName);
	}catch(RemoteException e){
	    agentDeath();
	}
    }
    boolean setProtocol(String _protocol){
	boolean res;
	synchronized(protocolLeaderMonitor){
	    res=setProtocolGUI(_protocol);
	}
	if(myGUI==null)
	    System.out.println("No myGUI");
	else
	    myGUI.setProtocol(_protocol);
	return res;
    }
    void tests(){
	try{
	    BrokerServer.oneWaySend(brokerServer,"TESTS",null);
	}catch(Exception e){
	    e.printStackTrace();
	    System.err.println("Broker:tests:sendTESTS");
	}
    }
    void tests_run(){
	testLeader.test();
    }
    boolean setProtocolGUI(String _protocol){
	if(_protocol.equals("_ABT_")){
	    System.out.println("Protocol..."+_protocol);
	    protocol=_protocol;
	    protocolLeader=(ProtocolLeader)new P_ABT_Leader(this);
	    return true;
	}
	// agrregations constrained on external
	if(_protocol.equals("C_ABT_")){
	    System.out.println("Protocol..."+_protocol);
	    protocol=_protocol;
	    protocolLeader=(ProtocolLeader)new P_ABT_Leader(this);
	    return true;
	}
	// aggregations unconstrained
	if(_protocol.equals("U_ABT_")){
	    System.out.println("Protocol..."+_protocol);
	    protocol=_protocol;
	    protocolLeader=(ProtocolLeader)new P_ABT_Leader(this);
	    return true;
	}
	// agregations constrained on internal
	if(_protocol.equals("I_ABT_")){
	    System.out.println("Protocol..."+_protocol);
	    protocol=_protocol;
	    protocolLeader=(ProtocolLeader)new P_ABT_Leader(this);
	    return true;
	}
	System.out.println("No protocol... "+_protocol);
	return false;
    }
    public void clearSession(String explication){
	setProtocol(defaultProtocol);
	System.out.println("Protocol is reset");
 	stopSession(explication);
	System.out.println("Session stop");
	String protocol=defaultProtocol;
	//IBrokerServer brokerServer;
	problemByAgent= new Hashtable();
	problemByVariable= new Hashtable();
	problemByAgentName= new Hashtable();
	//sortedAgents= new IAgentServer[0];
	//agentOrder= new Hashtable();
    }
    public void stopSession(String explication){
	debug=running=false;
	// no cycle !notifyAgents("STOPSEARCH",new DString(explication));
	
	IAgentServer agent;
	Iterator keys;
	keys = (Iterator) problemByAgent.keys();
	
	while (keys.hasNext()){
	    agent = (IAgentServer) keys.next();
	    try {
		BrokerServer.oneWaySend(agent,"STOPSEARCH",new DString(explication));
	    }
	    catch (Exception e){
		System.err.println("in Broker: StopSession");
		System.err.println("Agent "+((AgentData)problemByAgent.get(agent)).getName()+" was dead!");
		continue;
	    }
	}	
	
    }
    void resetProtocolLeader(){
	synchronized(protocolLeaderMonitor){
	    protocolLeader=null;
	}
    }
    void agentDeath(){
	testLeader.nrOfTests=0;
	try{
	    throw new Exception();
	}catch(Exception e){
	    endSolving("agent death..."+e);
	}
	clearSession("agent death...");
    }
    void setProtocol(){
	setProtocolGUI();
	if(myGUI!=null)
	    myGUI.setProtocol(defaultProtocol);
	else
	    System.out.println("No myGUI");
    }
    void setProtocolGUI(){
	synchronized(protocolLeaderMonitor){
	    System.out.println("Protocol..."+defaultProtocol);
	    protocol=defaultProtocol;
	    protocolLeader=new P_ABT_Leader(this);
	}
    }
    String getProtocol(){
	return protocol;
    }
    void notifyLeader(){
	synchronized (leaderSemaphore){
	    leaderSemaphore.inc();
	    leaderSemaphore.notify();
	}
    }
    void notifyAgents(String command, Object param){
	IAgentServer agent;
	Iterator keys;
	keys = (Iterator) problemByAgent.keys();
	
	while (keys.hasNext()){
	    agent = (IAgentServer) keys.next();
	    try {
		BrokerServer.oneWaySend(agent, command, param);
	    }
	    catch (RemoteException e){
		// At exit is normal
		if(command.equals("EXIT")) continue;
		//e.printStackTrace();
		System.err.println("In Broker: notify agent");
		System.err.println("Agent "+((AgentData)problemByAgent.get(agent)).getName()+" not answering! Killing session...");
		agentDeath();
		return;
	    }
	}	
    }   
    public void step(){
	debug=true;
	if(!running)
	    run();
	else{
	    notifyLeader();
	    notifyAgents("STEP",null);
	}
    }
    public void waitStep(){
	if(debug)
	    leaderSemaphore.requestWait();
    }
    public void exit(int cod){
	notifyAgents("EXIT","Administrator request");
	System.out.println("exit:Deciding to EXIT");
	System.err.println("exit:Deciding to EXIT");
	System.exit(cod);
    }
    public void run(){
	if(running){
	    System.err.println("Already running");
	    return;
	}
	running=true;
	try {
	    try {
		ProtocolLeader tmp=protocolLeader;
		synchronized(protocolLeaderMonitor){
		    if(tmp==null){
			System.err.println("run:No Leader!!");
			return;
		    }
		}
		tmp.run();
	    }catch (AgentDeathException e){
		agentDeath();
		System.err.println("Killing Done!");
	    }
	}
	catch (Exception e){
	    stopSession(e.getMessage());
	}
    }
    public String getName(){
	return name;
    }
    public boolean subscribe(IAgentServer agent, SVariable variables[], int importance, int complexity){
	int k;
	String agentName="";
	String varcrt;
	AgentData agentData=new AgentData();
	if (!agentData.set(agent,variables,importance,complexity)){
	    System.out.println("Subscription error");
	    return false;
	}
	agentName=agentData.getName();
	
	boolean resubscrib=false;
	if(problemByAgent.put(agent, agentData)!=null){
	    System.out.println("Resubscribing "+agentName);
	    resubscrib=true;
	}
	
	if(problemByAgentName.put(agentName, agent)!=null){
	    if(!resubscrib)
		System.out.println("Name conflict on: \""+agentName+"\"");		    
	}else{
	    if(resubscrib)
		System.out.println("\tThe resubscribtion was with new name: \""+agentName+"\"");
	}
	if(resubscrib){
	    try {
		eraseAgentFromVariables(agent);
	    }catch (Exception e){
		System.err.println("Error erasing agent "+agent+" from variables");
		clearSession("Broker internal error: Please announce application manager.");
		return false;
	    }
	}
	System.out.println("Subscribing :"+variables.length);
	for(k=0;k<variables.length;k++){
	    ArrayList alist;
	    varcrt=variables[k].getSVariable();
	    System.out.println("Add: "+agentName+ " "+ variables[k].getSVariable());
	    if(problemByVariable.containsKey(varcrt))
		alist=(ArrayList)problemByVariable.get(varcrt);
	    else
		alist= new ArrayList();
	    alist.add(new SAgent(agent));
	    problemByVariable.put(varcrt,alist);
	}
	return true;
    }
    
    void eraseAgentFromVariables(IAgentServer agent){
	int k,l,erasedNr;
	ArrayList[] alist;
	IAgentServer oldAgent;
	alist=(ArrayList[])(problemByVariable.values().toArray(new ArrayList[0]));
	erasedNr=l=0;
	for(k=0;k<alist.length;k++){
	    //System.out.println(erasedNr+" erased,variable "+k+" restartfrom "+l);
	    int alength;
	    alength=(alist[k]).size();
	    //System.out.println(erasedNr+" erased,variable "+k+" restartfrom "+restartFrom+" arraysize: "+alength);
	    for(l=0;l<alength;l++){
		//System.out.println(erasedNr+" erased,variable "+k+" restartfrom "+restartFrom+" arraysize: "+alength+" agentcrt: "+l);
		oldAgent=((SAgent)alist[k].get(l)).getSAgent();				    
		if(oldAgent.equals(agent)){
		    ((ArrayList)alist[k]).remove(l);
		    erasedNr++;
				//System.out.println("removing agent");
		    alength=(alist[k]).size();
		    l--;
		}
	    }
	}
	System.out.println(erasedNr+" already existing variables were erased");
    }
    public IAgentServer getByName(String name) throws RemoteException{
	IAgentServer agent;
	synchronized (problemByAgentName){
	    agent=(IAgentServer)problemByAgentName.get(name);
	}
	return agent;
    }
    public Hashtable getProblemByAgent(){
	return problemByAgent;
    }
    public Hashtable getProblemByVariable(){
	return problemByVariable;
    }
    public Hashtable getProblemByAgentName(){
	return problemByAgentName;
    }

    public void finalize() throws Throwable{
	System.out.println("broker dying!");
	super.finalize();
    }
}

