package dcsp;
import dcsp.parameters.*;
import java.rmi.*;
import java.util.Hashtable;
import java.util.*;

class P_ABT_Leader implements ProtocolLeader{
    Broker broker;
    //final String protocol="_ABT_";
    // cost buit to declare solution
    DCost acceptedCost=new DCost();
    // cost of messages
    final DCost messageCost=(new DCost()).setMessages();
    // search hash mapping IAgentServer to AgentData
    Hashtable problemByAgent;
    // search hash mapping String to ArrayList(SAgent)
    Hashtable problemByVariable;
    // search hash mapping String names to IAgentServer
    Hashtable problemByAgentName;
    //(IAgentServer, dAccepted)
    Hashtable accepted= new Hashtable();
    // endAgents (IAgentServer)
    ArrayList endAgents=new ArrayList();

    IAgentServer[] sortedAgents= new IAgentServer[0];
    // search hash mapping IAgentServer to position
    Hashtable agentOrder= new Hashtable();
    Object starting_links=new Object();


    P_ABT_Leader(Broker _broker){
	broker=_broker;
    }
    public void run() throws AgentDeathException, Exception{
	// to ensure a full run at early
	// EMPTYNOGOOD detection
	// or solution from multiple endAgents
	synchronized(starting_links){
	    synchroRun();
	}
    }
    private void printProblemByVariable(Hashtable problemByVariable){
	System.out.println("Problem by Variable:");
	if(problemByVariable==null){
	    System.out.println("Null problemByVariable");
	    return;
	}
	Iterator keys = (Iterator)problemByVariable.keys();
	while(keys.hasNext()){
	    String var=(String)keys.next();
	    System.out.print("Variable:"+var+" for the agents :");
	    ArrayList alist=(ArrayList)problemByVariable.get(var);
	    for(int k=0;k<alist.size();k++) {
		SAgent sa=(SAgent)alist.get(k);
		AgentData ad=(AgentData)problemByAgent.get(sa.getSAgent());
		if((ad!=null)&&(ad.name!=null))
		    System.out.print(ad.name+":");
	    }
	    System.out.println("");
	}
    }
    void synchroRun() throws AgentDeathException, Exception{
	problemByAgent=broker.getProblemByAgent();
	problemByVariable=broker.getProblemByVariable();
	problemByAgentName=broker.getProblemByAgentName();
	
	printProblemByVariable(problemByVariable);
	
	reOrder();
	IAgentServer agent;
	Iterator keys;
	keys = (Iterator) problemByAgent.keys();
	
	while (keys.hasNext()){
	    agent = (IAgentServer) keys.next();
	    DStartSearch params=new DStartSearch(broker.getProtocol());
	    params.position=((IntObject)agentOrder.get(agent)).getInt();
	    try {
		BrokerServer.oneWaySend(agent, "STARTSEARCH", params);
	    }catch (RemoteException e){
		System.err.println("Agent "+((AgentData)problemByAgent.get(agent)).getName()+" not answering! Killing session...");
		throw new AgentDeathException(e.getMessage());
	    }
	    DIncomingAgent incoming=new DIncomingAgent(getInComing(agent));
	    if(incoming.get().length==0){
		endAgents.add(agent);
		System.out.println("Agent "+((AgentData)problemByAgent.get(agent)).getName()+" is endAgent");
	    }
	    DOutgoingAgent outgoing=new DOutgoingAgent(getOutGoing(agent));
	    DSearch dSearch=new DSearch();
	    dSearch.agentOrder=agentOrder;
	    dSearch.sortedAgents=(Object)sortedAgents;
	    dSearch.problemByAgent=problemByAgent;
	    DStartLinks dStartLinks=new DStartLinks();
	    dStartLinks.dIncomingAgent=incoming;
	    dStartLinks.dOutgoingAgent=outgoing;
	    dStartLinks.dSearch=dSearch;
	    try {
		BrokerServer.oneWaySend(agent, "START_LINKS",dStartLinks);
	    }catch (Exception e) {
		e.printStackTrace();
		System.err.println("Error register:");
		System.err.println("    Agent \""+((AgentData)problemByAgent.get(agent)).getName()+"\" cannot get outgoing links");
		throw new AgentDeathException("Agent death at getting outgoing");
	    }
	}
    }
    protected void finalize() throws Throwable{
	Utils.runFinal("P_ABT_Leader");
    }
    public Object message(String command, Object parameters){
	if(command.equals("GETLAST")){
	    getLast((DGetLast)parameters);
	    return null;
	}	
	if(command.equals("GETANTECEDENTS")){
	    return getAntecedents((DGetAntecedents)parameters);
	}
	// to ensure a full run at early
	// EMPTYNOGOOD detection
	if(command.equals("EMPTYNOGOOD")){
	    synchronized(starting_links){
		broker.clearSession("Empty nogood");
	    }
	    broker.endSolving("Empty nogood:\n"+parameters);
	    return null;
	}	
	if(command.equals("ACCEPTED")){
	    //System.out.println("The end occurs when all the incoming intersect nonempty.\n"+parameters.toString());
	    synchronized(starting_links){
		receiveAccepted((DAccepted)parameters);
	    }
	    return null;
	}
	if(command.equals("REMOVE_END_AGENT")){
	    synchronized(starting_links){
		removeEndAgent((IAgentServer)parameters);
	    }
	    return null;
	}
	System.err.println("Unknown command "+command);
	return null;
    }
    // true if end of search
    private boolean receiveAccepted(DAccepted daccepted){
	accepted.put(daccepted.agent,daccepted);
	acceptedCost.updateCost(daccepted.cost);
	if(accepted.size()==endAgents.size())
	    return testEndOfSearch();
	return false;
    }
    private boolean testEndOfSearch(){
	if(accepted.isEmpty()){
	    System.err.println("End of search, but empty!");
	    broker.clearSession("No end Agents!!");
	    broker.endSolving("No end Agents!!");
	    return true;
	}
	DAccepted dAccepted=new DAccepted();
	//dAccepted.agent=broker.brokerServer;
	dAccepted.history=IntersectionOfSolutions();
	if(dAccepted.empty()&&
	   (!dAccepted.vide())){
	    System.err.println("Empty Accepted!!");
	    return false;
	}
	System.out.println(dAccepted.toString());
	System.out.println(acceptedCost.toString());
	broker.clearSession("Solved!");
	broker.endSolving(dAccepted.toString()+"\n"+acceptedCost.toString()+"\n");
	return true;
    }
    private Hashtable IntersectionOfSolutions(){
	Hashtable view=new Hashtable();
	Iterator it=(Iterator)accepted.keys();
	while(it.hasNext()){
	    System.out.println("Adding solution from accepted");
	    addSolution(view,((DAccepted)accepted.get(it.next())).history);
	}
	return view;
    }
    // solution=(String,DVarHistory)
    private void addSolution(Hashtable view, Hashtable solution){
	DVarHistory solHist, accHist;
	Iterator it=(Iterator) solution.keys();
	while(it.hasNext()){
	    String var=(String)it.next();
	    solHist=(DVarHistory)view.get(var);
	    accHist=(DVarHistory)solution.get(var);
	    if(solHist==null)
		view.put(var,accHist.clone());
	    else{
		solHist.restrict(accHist);
	    }
	}
    }
    void removeEndAgent(IAgentServer agent){
	int pos=endAgents.indexOf(agent);
	if(pos<0)
	    System.err.println("Agent not found for removing while pos<0");
	else try{
	    endAgents.remove(pos);
	    accepted.remove(agent);
	}catch(IndexOutOfBoundsException e){
	    System.err.println("Agent not found for removing, Index out of bound exception");
	}
	if(accepted.size()==endAgents.size())
	    testEndOfSearch();
    }
    IAgentServer[] getAntecedents(DGetAntecedents antecedents){
	ArrayList agents=new ArrayList();
	int id=antecedents.id;
	for(int k=id-1;k>=0;k--)
	    if(hasAny(sortedAgents[k],antecedents.variables)){
		agents.add(sortedAgents[k]);
		System.out.println("antecedent agent "+k);
	    }
	System.out.println("GetAntecedents");
	return (IAgentServer[])agents.toArray(new IAgentServer[0]);
    }
    void getLast(DGetLast request){
	IAgentServer target=(IAgentServer)request.source;
	int crtIndex=((IntObject)agentOrder.get(target)).getInt();
	System.out.println("Index of source agent: "+crtIndex);
	try{
	    for(int k=crtIndex-1;k>=0;k--)
		if(hasAny(sortedAgents[k],request.guiltyV)){
		    System.out.println("Found target at position: "+k);
		    BrokerServer.oneWaySend(target,"LAST",sortedAgents[k]);
		    return;
		}
	    BrokerServer.oneWaySend(target, "LAST",null);
	}catch(RemoteException e) {
	    broker.notifyAgents("EXIT","dead agent");
	    System.err.println("EXIT-dead agent");
	    System.exit(1);
	}
    }
    boolean hasAny(IAgentServer agent, String[] variables){
	SVariable[] svars=((AgentData)problemByAgent.get(agent)).getSVariables();
	for(int k=0;k<svars.length;k++)
	    for(int kk=0;kk<variables.length;kk++)
		if(variables[kk].equals(svars[k].getSVariable()))
		    return true;
	return false;
    }
    void sort(AgentData[] ad){
	// sort by importance
	//forall with same importance{
	//  sort by owner
	//     forall with same importance&&owner
	//        sort by complexity
	try {
	    Arrays.sort(ad,new AgentDataImportanceComparator());
	}catch (Exception e){
	    System.err.println("Error while sorting agents by importance");
	}
	// for all subsets with equal cost
	int upperToConsider,lowerToConsider=0;
	while(lowerToConsider<ad.length){
	    int k=lowerToConsider;
	    //System.out.println("sort:lowerToConsider="+k);
	    int crtImportance=ad[k].getImportance();
	    for(k++;;k++){
		if(k>=ad.length)
		    break;
		if(ad[k].getImportance()!=crtImportance)
		    break;
	    }
	    upperToConsider=k-1;
	    //System.out.println("sort:=upperToConsider"+upperToConsider);
	    // building ArrayList parameter
	    ArrayList ordered=new ArrayList(upperToConsider-lowerToConsider+1);
	    for(k=lowerToConsider;k<=upperToConsider;k++)
		ordered.add(ad[k]);
	    AgentData add[];
	    add=sortWithCycles(ordered);
	    // got order
	    for(k=lowerToConsider;k<=upperToConsider;k++){
		ad[k]=add[k-lowerToConsider];
		if(ad[k]==null)
		    System.err.println("sort:null");
	    }
	    lowerToConsider=upperToConsider+1;
	}
	//System.out.println("sort:End");
    }
    private void resetMarks(AgentGraph[] ag, int kk){
	for(int k=0;k<=kk;k++){
	    if(ag[k]==null){
		System.err.println("resetMarks:null:"+k);
	    }else
		ag[k].mark=false;
	}
    }
    AgentData[] sortWithCycles(ArrayList al){
	int k;
	AgentGraph ag[]=new AgentGraph[al.size()];
	//System.out.println("sortWithCycles:size="+ag.length);
	for(k=0;k<ag.length;k++){
	    //System.out.println("sortWithCycles:k="+k);
	    AgentData ad=(AgentData)al.get(k);
	    ag[k]=new AgentGraph(ad);
	    //System.out.println("sortWithCycles:ad="+ad.name);
	    for(int kk=0;kk<k;kk++){
		//System.out.println("sortWithCycles:deal with="+ag[kk].ad.name);
		//those we can inferr to be after
		// if cycles, then this brings suboptimalities
		if(ag[kk].mark){
		    //System.out.println("sortWithCycles:marked="+ag[kk].ad.name);
		    continue;
		}
		if(AgentDataComparator.isBefore(ad,ag[kk].getAD())){
		    //System.out.println("sortWithCycles:before:"+ag[k].ad.name+" to:"+ag[kk].ad.name);
		    ag[k].addLink(ag[kk]);
		    // by transitivity, all the followers follow
		    ag[kk].markFollowers();
		}
		if(AgentDataComparator.isBefore(ag[kk].getAD(),ad)){
		    //System.out.println("sortWithCycles:before:"+ag[kk].ad.name+" to:"+ag[k].ad.name);
		    ag[kk].addLink(ag[k]);
		}
	    }
	    resetMarks(ag,k);
	}
	ArrayList competitors=new ArrayList();
	// get the candidates to first
	for(k=0;k<ag.length;k++){
	    // if no followers
	    if(ag[k].first()){
		//System.out.println("sortWithCycles:among first competitors:"+ag[k].ad.name);
		competitors.add(ag[k]);
		ag[k].mark=true;
	    }
	}
	//System.out.println("sortWithCycles:start sorting");
	// the sorting
	AgentData result[]=new AgentData[ag.length];
	for(k=0;k<ag.length;k++){
	    if(competitors.size()==0){
		System.out.println("Cycle !!");
		for(int kk=0;kk<ag.length;kk++){
		    if(ag[kk].mark) continue;
		    ag[kk].mark=true;
		    competitors.add(ag[kk]);
		    System.out.println("sortWithCycles:Cycle broken at:"+ag[kk].ad.name);
		    break;
		}
		if((competitors.size()==0)){
		    System.err.println("sortWithCycles:Cycle not broken!!");
		    break;
		}
	    }
	    AgentGraph[] tmp=(AgentGraph[])competitors.toArray(new AgentGraph[0]);
	    Arrays.sort(tmp,new AgentGraphComplexityComparator());
	    if(tmp==null){
		System.err.println("sortWithCycles:null tmp!!");
		continue;
	    }
	    if(tmp.length==0){
		System.err.print("sortWithCycles:empty tmp!!");
		continue;
	    }
	    //System.out.print("sortWithCycles:Sorted: ");
	    for(int kk=0;kk<tmp.length;kk++){
		if(tmp[kk]==null){
		    System.err.println("sortWithCycles:tmp["+kk+"]=null!!");
		    continue;
		}
		//System.out.print(tmp[kk].ad.name+" ");
	    }
	    //System.out.println("");
	    result[k]=tmp[0].getAD();
	    //erase links and adds to competitors those that are not marked
	    //System.out.println("sortWithCycles:freeFollowers of:"+tmp[0].ad.name);
	    tmp[0].freeFollowers(competitors);
	    int pos=competitors.indexOf(tmp[0]);
	    if(pos<0){
		System.err.println("Not found:"+tmp[0].ad.name+" "+pos);
	    }else{
		//System.out.println("sortWithCycles:remove"+tmp[0].ad.name+" pos:"+pos);
		competitors.remove(pos);
	    }
	}
	System.out.print("Obtained: ");
	for(int kk=0;kk<result.length;kk++){
	    System.out.print(result[kk].name+" ");
	}
	System.out.println("");
	return result;
    }
    void reOrder(){
	AgentData[] ad;
	ad=(AgentData[]) problemByAgent.values().toArray(new AgentData[0]);

	sort(ad);
	int agentsNr=ad.length;
	System.out.println("Subscribed "+agentsNr+" agent(s)!");
	agentOrder=new Hashtable(agentsNr);
	sortedAgents= new IAgentServer[agentsNr];
	for(int i=0;i<agentsNr;i++){
	    sortedAgents[i]=ad[i].getAgent();
	    IntObject io=new IntObject(i);
	    agentOrder.put(sortedAgents[i],io);
	    System.out.println("Agent "+ad[i].getName()+" has position "+i);
	}
    }
    OutgoingAgent[] getOutGoing(IAgentServer agent){
	OutgoingAgent oAgent[]=new OutgoingAgent[0];
	Hashtable outgoing=new Hashtable();
	SVariable svars[]=new SVariable[0];
	int pos1,pos2;
	IntObject i1,i2;
	i1=(IntObject)agentOrder.get(agent);
	pos1=i1.getInt();
	AgentData ad=(AgentData) problemByAgent.get(agent);
	if (ad!=null)
	    svars=ad.getSVariables();
	else 
	    return oAgent;
	int svarsNr=svars.length;
	for(int k=0;k<svarsNr;k++){
	    try {
		String variableCrt=svars[k].getSVariable();
		ArrayList alist;
		alist=(ArrayList)problemByVariable.get(variableCrt);
		int vAgentsNr=alist.size();
		for(int a=0;a<vAgentsNr;a++){
		    SAgent sa=(SAgent) alist.get(a);
		    IAgentServer ag=sa.getSAgent();
		    i2=(IntObject)agentOrder.get(ag);
		    pos2=i2.getInt();
		    if(pos1<pos2){
			System.out.println("\tVariable "+variableCrt+" agent "+((AgentData)problemByAgent.get(agent)).getName()+" has position "+pos1+" while "+((AgentData)problemByAgent.get(ag)).getName()+" has position "+pos2);
			OutgoingAgent oa;
			oa=(OutgoingAgent)outgoing.get(ag); 
			//System.out.println("step1 ");
			if(oa==null){
			    oa=new OutgoingAgent(ag);
			    //System.out.println("step2 ");
			}
			else{
			    System.out.println("\tAlready existing link from "+((AgentData)problemByAgent.get(agent)).getName()+" to "+((AgentData)problemByAgent.get(ag)).getName());
			}
			oa.insert(variableCrt);
			outgoing.put(ag,oa);
		    }
		}
	    }
	    catch (Exception e){
		System.err.println("P_ABT_Leader:getOutgoing:Error getting agents list for variable "+svars[k].getSVariable());
		System.err.println(e);
		e.printStackTrace();
	    }
	}
	try {
	    oAgent= (OutgoingAgent[]) outgoing.values().toArray(new OutgoingAgent[0]);
	}
	catch (Exception e){
	    System.err.println("P_ABT_Leader:Error getting list of outgoing");
	    System.err.println(e);
	    e.printStackTrace();
	}
	return oAgent;
    }
    IncomingAgent[] getInComing(IAgentServer agent){
	IncomingAgent iAgent[]=new IncomingAgent[0];
	Hashtable incoming=new Hashtable();
	SVariable svars[]=new SVariable[0];
	int pos1,pos2;
	IntObject i1,i2;
	i1=(IntObject)agentOrder.get(agent);
	pos1=i1.getInt();
	AgentData ad=(AgentData) problemByAgent.get(agent);
	if (ad!=null)
	    svars=ad.getSVariables();
	else
	    return iAgent;
	int svarsNr=svars.length;
	for(int k=0;k<svarsNr;k++){
	    try {
		String variableCrt=svars[k].getSVariable();
		ArrayList alist;
		alist=(ArrayList)problemByVariable.get(variableCrt);
		int vAgentsNr=alist.size();
		for(int a=0;a<vAgentsNr;a++){
		    SAgent sa=(SAgent) alist.get(a);
		    IAgentServer ag=sa.getSAgent();
		    i2=(IntObject)agentOrder.get(ag);
		    pos2=i2.getInt();
		    if(pos1>pos2){
			IncomingAgent oa;
			oa=(IncomingAgent)incoming.get(ag); 
			if(oa==null)
			    oa=new IncomingAgent(ag);
			incoming.put(ag,oa);
		    }
		}
	    } catch (Exception e){
		e.printStackTrace();
	    }
	}
	try {
	    iAgent= (IncomingAgent[]) incoming.values().toArray(new IncomingAgent[0]);
	}
	catch (Exception e){
	    System.err.println("P_ABT_Leader"+e);
	    e.printStackTrace();
	}
	return iAgent;
    }
}

