package dcsp;
import dcsp.parameters.*;
import java.io.*;
import java.rmi.*;

class Problem {
    Agent owner;
    SetOfConstraints cset;
    SetOfVariables vset;
    Solver solver=new Solver();
    // allows/forbids the access to the solver
    boolean setup=false;
    // monitor of the pointer to solver
    // for the release of the solver
    Object setupMonitor=new Object();
    String crtProtocol;
    Problem(Agent _owner){
	owner=_owner;
	vset=new SetOfVariables(this);
	cset=new SetOfConstraints(this);
    }
    public int getVariablesNr(){
	return vset.getVariablesNr();
    }
    public Object message(String command, Object parameters){
	if(command.equals("STARTSEARCH")){
	    try {
		// modifies the pointer to solver
		System.out.println(command+" waits for solverMonitor:id2");
		synchronized(setupMonitor){
		    startSearch(((DStartSearch)parameters).get(),parameters);
		    setup=true;
		    setupMonitor.notifyAll();
		}
	    }catch (Exception e){
		e.printStackTrace();
		try {
		    AgentServer.oneWaySend(owner.broker,"PROTOCOL ERROR", new DString(owner.getName()));
		}catch (RemoteException evt){
		    System.err.println("Dead connection to broker at Protocol Error");
		}
	    }
	    return new DOk();
	}
	
	System.out.println(command+" waits for setupMonitor:id2");
	synchronized(setupMonitor){
	    try {
		while(!setup)
		    setupMonitor.wait();
	    }catch (Exception e){
		System.err.println(command+" interrupted while waiting setupMonitor->exit");
		return new DOk();
	    }
	}
	if(command.equals("STOPSEARCH")){
	    Solver s;
	    System.out.println(command+" waits for solverMonitor:id3");
	    synchronized(setupMonitor){
		setup=false;
		if(solver==null){
		    System.out.println(command+" abort because Empty solver");
		    return null;
		}
		s=solver;
		solver=null;
	    }
	    return s.message(command, parameters);
	}
	Solver tmp;
	Object result;
	System.out.println(command+" waits for solverMonitor:id1");
	synchronized(setupMonitor){
	    if(solver==null){
		System.out.println(command+" abort because Empty solver");
		return null;
	    }
	    tmp=solver;
	}
	result=tmp.message(command, parameters);
	return result;
    }
    public Agent getAgent(){
	return owner;
    }
    public SetOfVariables getSetOfVariables(){
	return vset;
    }
    public SVariable[] getSVariables(){
	return getSetOfVariables().getSVariables();
    }
    public int getComplexity(){
	return cset.getComplexity();
    }
    public SetOfConstraints getSetOfConstraints(){
	return cset;
    }
    public void startSearch(String protocol, Object params) throws Exception{
	crtProtocol=protocol;
	Utils.aggregations=false;
	Utils.constrained=false;
	Utils.inversConstrained=false;
	Utils.nogoodToLastFirst=false;
	// sending OK? only for owned
	Utils.yokoo=false;
	// first iff owned
	Utils.firstOwned=true;
	if(protocol.equals("_ABT_")){
	    solver=(Solver)new _ABT_Solver(this);
	    solver.setParams(params);
	    Utils.aggregations=false;
	    Utils.constrained=false;
	    Utils.inversConstrained=false;
	    Utils.nogoodToLastFirst=true;
	    Utils.yokoo=true;
	    System.out.println("Protocol \""+protocol+"\" loaded");
	    solver.Solve();
	    return;
	}
	// agrregations constrained on external
	if(protocol.equals("C_ABT_")){
	    solver=(Solver)new _ABT_Solver(this);
	    solver.setParams(params);
	    Utils.aggregations=true;
	    Utils.constrained=true;
	    Utils.inversConstrained=false;
	    System.out.println("Protocol \""+protocol+"\" loaded");
	    solver.Solve();
	    return;
	}
	// aggregations unconstrained
	if(protocol.equals("U_ABT_")){
	    solver=(Solver)new _ABT_Solver(this);
	    solver.setParams(params);
	    Utils.aggregations=true;
	    Utils.constrained=false;
	    Utils.inversConstrained=false;
	    System.out.println("Protocol \""+protocol+"\" loaded");
	    solver.Solve();
	    return;
	}
	// agregations constrained on internal
	if(protocol.equals("I_ABT_")){
	    solver=(Solver)new _ABT_Solver(this);
	    solver.setParams(params);
	    Utils.aggregations=true;
	    Utils.constrained=true;
	    Utils.inversConstrained=true;
	    System.out.println("Protocol \""+protocol+"\" loaded");
	    solver.Solve();
	    return;
	}
	throw new Exception("protocol "+protocol+" absent");
    }
    
    public boolean loadFile(File file){
	String buf = "";
	RandomAccessFile fis=null;
	try {
	    fis= new RandomAccessFile(file, "r");
	    buf=Utils.getString(fis);
	}catch (Exception e){
	    System.err.println("Problem:loadFile:"+e);
	    e.printStackTrace();
	    try{
		fis.close();
	    }catch(java.io.IOException ev){
		System.err.println(ev);
		ev.printStackTrace();
	    }
	    return false;
	}
	if (buf.equals("COLOR")){
	    System.out.println("Loading COLOR ");
	    colorLoader(file,fis,0);
	    try{
		fis.close();
	    }catch(Exception ev){
		System.err.println(ev);
		ev.printStackTrace();
	    }
	    return true;
	}
	if (buf.equals("COLOR_V1")){
	    System.out.println("Loading COLOR_V1 ");
	    colorLoader(file,fis,1);
	    try{
		fis.close();
	    }catch(Exception ev){
		System.err.println(ev);
		ev.printStackTrace();
	    }
	    return true;
	}	
	System.out.println("Unknown file Type:\""+buf+"\"");
	if(fis!=null)
	    try{
		fis.close();
	    }catch(Exception ev){
		System.err.println(ev);
		ev.printStackTrace();
	    }
	return false;
    }
    void cmp(String s1, String s2){
	byte b1[]=s1.getBytes();
	byte b2[]=s2.getBytes();
	int k;
	System.out.println("Comparing "+s1+" to "+s2);
	for(k=0;k<b1.length; k++)
	    System.out.print(b1[k]+":");
	System.out.println("EOL");
	for(k=0;k<b2.length; k++)
	    System.out.print(b2[k]+":");
	System.out.println("EOL");
    }

    /*
    void colorLoader(File file, RandomAccessFile fis)
    {
	int constraintsNr=Utils.getInt(fis);
	System.out.println(constraintsNr +" Constraints");
	int k=0;
	for(;k<constraintsNr; k++)
	    {
		int varNr=Utils.getInt(fis);
		//vset.colorLoader(fis,varNr);
		//Reads the domain Size from vset
		cset.colorLoader(fis,varNr,file.getName()+"://"+k);
	    }
    }
    */
    void colorLoader(File file, RandomAccessFile fis, int fileType){
	int constraintsNr=Utils.getInt(fis);
	System.out.println(constraintsNr +" Constraints");
	int k=0;
	for(;k<constraintsNr; k++){
		int varNr=Utils.getInt(fis);
		//vset.colorLoader(fis,varNr);
		//Reads the domain Size from vset
		cset.colorLoader(fis,varNr,file.getName()+"://"+k, fileType);
	    }
    }
}
