package dcsp;
import dcsp.parameters.*;

class Scanable {
    int dimensions;
    int[] startPoint;
    Range[] range;
    int[] order; // first most important
    int[] variable; // first the index of the first variable
    boolean orderSet=false;
    Scanable(Feasible region){
        range=region.get();

        dimensions=range.length;
        startPoint=new int[dimensions];
        order=new int[dimensions];
        variable=new int[dimensions];
        for(int k=0;k<dimensions;k++){
	    order[k]=variable[k]=k;
	    startPoint[k]=range[k].getL();
	}
    }
    Scanable(int _dimensions){
	setDimension(_dimensions);
    }
    private void setDimension(int _dimensions){
        dimensions=_dimensions;
        startPoint=new int[dimensions];
        order=new int[dimensions];
        variable=new int[dimensions];
        range=new Range[dimensions];
        for(int k=0;k<dimensions;k++){
            order[k]=variable[k]=k;
            startPoint[k]=0;
        }
    }
    public void reset(){
        System.out.println("Reset startPoint");
        for(int k=0;k<dimensions;k++){
            //order[k]=variable[k]=k;
            startPoint[k]=range[k].getL();
        }       
        orderSet=false;
    }
    public Scanable setDim(int[] dim){
	setDimension(dim.length);
        for(int k=0;k<dimensions;k++){
	    System.out.println("Wanna dimension..."+k);
            range[k]=new Range(0,dim[k]-1);
            startPoint[k]=range[k].getL();
        }
	System.out.println("Wanna print...");
        Print();
        return this;
    }
    public void setOrder(int[] _order, int[] _invorder){
        if(!orderSet){
	    // not for unidimensional
	    if(_order.length>1){
		System.out.println("Setting fixed order ["+_order[0]+","+ _order[1]+" ]");
	    }
	    //orderSet=true;
	    order=(int[])_order.clone();
	    variable=(int[])_invorder.clone();
	}
        else
            System.out.println("Fixed order not changed");
	
    }
    void regionPrint(){
        for(int k=0;k<dimensions;k++){
            System.out.println("["+range[k].getL()+"\t,"+range[k].getU()+"]");
        }        
    }
    public void Print(){
        System.out.print("Start points: ");
        for(int k=0;k<dimensions;k++){
            System.out.print(startPoint[k]+" ");
        }
        System.out.println("");
        regionPrint();
    }
    // order0 is the direction on which we have tried surely to expand
    // if aggregations were enabled
    public void mark(Range[] _range, int order0){ 
        orderSet=true; //used for marking start Point
        // if we ever want to reduce scanables
        //range[order[dimensions-1]].setL(_range[order[dimensions-1]].getL());
        int k;
        System.out.print("Old start points: ");
        for(k=0;k<dimensions;k++){
            System.out.print(startPoint[k]+" ");
        }
        for(k=1;k<dimensions;k++){
            startPoint[order[k]]=_range[order[k]].getL();
        }
        // one after was already verified while the other will
        // be the first one
	if(Utils.aggregations && (order0==order[0]))
	    startPoint[order[0]]=_range[order[0]].getU()+2;
	else
	    startPoint[order[0]]=_range[order[0]].getU()+1;
	System.out.print("\nNew start points: ");
        for(k=0;k<dimensions;k++){
            System.out.print(startPoint[k]+" ");
        }
        System.out.println("");
    }
    public int getStartPoint(int index){
        return startPoint[index];
    }
    public int[] getOrder(){
        orderSet=true; // used in exterior
        return order;
    }
    public int[] getReorder(){
        orderSet=true; // used in exterior
        return variable;
    }
    public Range restrict(Range _range, int index){
        _range.restrict(range[index]);
        //_range.restrictL(range[index].getL());
        //_range.restrictU(range[index].getU());  
        return _range;
    }
}
