package dcsp;
import java.util.*;
import dcsp.parameters.*;

class YConstraint implements IYConstraint {
    CRepresentation mat;
    Cover cover;
    Scanable crtRegion; // last toScan region
    boolean crtRegionActive=false;
    ArrayList toScan, nogoods, feasible;
    int[] varIDs;
    // first most important
    int[] order; // the recommended order for this search 
    // first the index of the first variable
    int[] invorder; // invorder for this search 
    int[] score; //scores for ordering(hard ordered)

    _ABT_SolverThread yThread;

    YConstraint(_ABT_SolverThread _yThread){
        yThread=_yThread;
	init_order(0);
    }
    YConstraint(Constraint c, _ABT_SolverThread _yThread){
        yThread=_yThread;
        assocConstraint(c); //init. mat
	int varNr=getVariablesNr();
	System.out.println("Constraint "+varNr);
	switch(varNr){
	case 1:
	    cover=(Cover) new Cover1((C1Representation)mat);
	    toScan=new ArrayList();
	    nogoods=new ArrayList();
	    feasible=new ArrayList();
	    toScan.add((new Scanable(1)).setDim(mat.getDim()));
	    break;
	case 2:
	    cover=(Cover) new Cover2((C2Representation)mat);
	    toScan=new ArrayList();
	    nogoods=new ArrayList();
	    feasible=new ArrayList();
	    toScan.add((new Scanable(2)).setDim(mat.getDim()));
	    break;
	default:
	    System.err.println("EXIT-Illegal constraint arity:"+varNr);
	    System.exit(1);
	}
	init_order(varNr);
    }
    void init_order(int dim){
	if(dim==0)
	    return;
	int _order[]=new int[dim];
	int _score[]=new int[dim];
	int _invorder[]=new int[dim];
	for(int k=0;k<dim;k++){
	    _score[k]=0;
	    _order[k]=invorder[k]=k;
	}
	setOrder(_score, _order, _invorder);
    }
    // initializ mat,order,(memory varIDs)
    void assocConstraint(Constraint c){
        int nrVars;
        mat=c.getRepresentation();
        // needs mat
        varIDs=new int[nrVars=getVariablesNr()];
	// order will be reset at the end
	// of the constructor
        order=new int[nrVars];
        invorder=new int[nrVars];
        score=new int[nrVars];
        for(int k=0;k<nrVars;k++){
            invorder[k]=order[k]=k;
            score[k]=0;
        }
    }
    // existance tested before call
    void loadFeasible(Range[] range){
        //if(feasible.isEmpty());
        Feasible crtfRegion;
        try {
            crtfRegion=(Feasible)feasible.remove(0);
        Range[] _range=crtfRegion.get();
        for(int k=0;k<range.length;k++)
            range[k]=(Range)_range[k].clone();
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
    void EmptyFeasible(){
        while(!feasible.isEmpty()){
            toScan.add(new Scanable((Feasible)feasible.remove(0)));
            }
    }
    boolean getFirstBox(Range[] result, Range[] domain){
        if(!feasible.isEmpty()){
            loadFeasible(result);
            return true;
        }
        
        if(toScan.isEmpty())
            return false;
      
        do{
            crtRegion=(Scanable)toScan.remove(0);
            System.out.print("Dealing with :");
            crtRegion.Print();
            if(cover.findIn(result,domain,crtRegion)){
                toScan.add(0,crtRegion);
                return crtRegionActive=true;
            }
        } 
        while(!toScan.isEmpty());
        return false;
    }
    void expand(Range[] box, Range[] domain){
	boolean excess=false;
        cover.setScore(score);
        for(int k=0;k<order.length;k++){ 
	    if(Utils.constrained){
		//first with excess not bothering
		excess=(k!=0);
	    }
            cover.expand(box,domain,order[k],excess);
        }
    }
    public boolean haveID(int id){
        for(int k=0; k<varIDs.length;k++)
            if(varIDs[k]==id)
                return true;
        return false;
    }
    void PrintBox(Range[] range, String comment){
        if(yThread.TRACE_ON){
            System.out.println(comment);
            System.out.println(""+range.length+"="+varIDs.length);
            for(int k=0;k<varIDs.length;k++)
                System.out.println(yThread.getVariable(getID(k))+"\t=[ "+ range[k].getL()+"\t, "+range[k].getU()+"\t]");
            System.out.println();
        }
    }
    private void initScanables(){
        toScan=new ArrayList();
        nogoods=new ArrayList();
        feasible=new ArrayList();
        toScan.add((new Scanable(2)).setDim(mat.getDim()));
    }
    private void resetScanables(){
        toScan=new ArrayList();
        nogoods=new ArrayList();
        feasible=new ArrayList();
	Scanable sc=new Scanable(2);
	sc.setDim(mat.getDim());
	sc.setOrder(order,invorder);
        toScan.add(sc);
    }
    public void initCover(){
        System.out.println("init Cover");
        cover.init(mat); //init of the matrix
        initScanables();
        resetContinuationPoints();
    }
    // Reset the Cover to only this box
    public void resetMark(Range[] box){
        System.out.println("init Cover");
        cover.init(mat); //init of the matrix
        resetScanables();
        cover.mark(box);
	PrintBox(box,"set Mark on:");
    }
    // Just reset of the scanables without reseting the marked map.
    public void resetContinuationPoints(){
        Iterator iter=(Iterator)toScan.iterator();
        Scanable crtRegionIt;
        while(iter.hasNext()){
            crtRegionIt=(Scanable)iter.next();
            crtRegionIt.reset();
        }
    }
    public void Print(){
	try{
	    System.out.println("Wanna print cover ");
	    cover.Print();
	    System.out.println("Wanna print scan ");
	    if(!toScan.isEmpty()){
		((Scanable)toScan.get(0)).Print();
	    }else{
		System.out.println("Empty toScan");
	    }
	} catch (IllegalMonitorStateException e) {
	    // a bug of java
	    e.printStackTrace();
	    System.err.println("YConstraint:Print:IllegalMonitorStateException");
	}
    }
    public boolean getNextBox(Range[] result, Range[] domain){
        System.out.println("getNextBox:");
        Print();
        PrintBox(domain,"Search in domain:");
        if(!getFirstBox(result,domain))
            return false;

        PrintBox(result,"First found:");
	if(Utils.aggregations){
	    //aggregations
	    expand(result, domain);
	    PrintBox(result,"Expanded to:");
	}
	System.out.println("Cover before marking");
	cover.Print();
	cover.mark(result);
	
	System.out.println("Cover after marking");
        cover.Print();
        
        if(crtRegionActive){
            crtRegionActive=false;
            crtRegion.mark(result,order[0]);
            crtRegion=null;
        }
        EmptyFeasible();
        return true;
    }
    public int getVariablesNr(){
        return mat.getVariablesNr();
    }
    public String getVariable(int index){
        return mat.getVariable(index);
    }
    public void setID(int index,int id){
        varIDs[index]=id;
    }
    public void setOrder(int[] _score, int[] _order, int[] _invorder){
	// The one directional constraints don't need
	if(getVariablesNr()>1){
	    System.out.println("Setting order ["+_order[0]+","+ _order[1]+" ]");
	    System.out.println("Setting score ["+_score[0]+","+ _score[1]+" ]");
	}
        score=_score;
        order=_order;
        invorder=_invorder;

        Iterator iter=(Iterator)toScan.iterator();
        Scanable crtRegionIt;
        while(iter.hasNext()){
            crtRegionIt=(Scanable)iter.next();
            crtRegionIt.setOrder(_order,_invorder);
        }       
    }
    public int getID(int index){
        return varIDs[index];
    }   
    public CRepresentation getRepresentation(){
        return mat;
    }
    // only for nogoods
    public void initDomain(Range[]domain ){
    }
}
