package dcsp;
import java.util.*;
import dcsp.parameters.*;

// domain -3 = - infinity
// domain -2 = + infinity

class YNogood implements IYConstraint {
    _ABT_SolverThread process;
    _ABT_Solver solver;
    //final int INF=-3;
    // the incoming information
    DNogood dNogood;
    Hashtable view;
    DVarHistory[] dVarHistory;// extends Range
    // the whole space
    Range[] forbid;
    String[] variables;
    //  global ids
    int[] varIDs;
    // nr of variables
    int varNr;
    boolean intersects=false;
    // A list to keep the splitted domains obtained
    ArrayList cover=new ArrayList();
    YNogood(DNogood dnogood, _ABT_SolverThread _process){
        process=_process;
	solver=process.getSolver();
        dNogood=dnogood;
        view=dNogood.view;
        dVarHistory=(DVarHistory[])(view.values()).toArray(new DVarHistory[0]);
        variables=(String[])view.keySet().toArray(new String[0]);
        varNr=variables.length;
        forbid=new Range[varNr];
        varIDs=new int[varNr];
        for(int k=0;k<varNr;k++){
            forbid[k]=(Range)view.get(variables[k]);
        }
    }
    // returns 0 if it should remain active
    //         1 partially active
    //         -1 inactive
    // parameter (String:DVarHistory)
    int active(Hashtable view){
	int result=0; //active
	for(int k=0;k<varNr;k++){
	    String crtVariable=variables[k];
	    DVarHistory vHist=(DVarHistory)view.get(crtVariable);
	    // normally here we should check the
	    // range from solver(*), perhaps they
	    // are partiall
	    if(vHist==null) continue;
	    // completely outside
	    if((vHist.getL()>forbid[k].getU())
	       ||(vHist.getU()<forbid[k].getL()))
		return -1;
	    // internal -no matter if partial or full
	    if(varIDs[k]>=0)
		continue;
	    // inside
	    if((vHist.getL()>=forbid[k].getL())
	       &&(vHist.getU()<=forbid[k].getU()))
		continue;
	    // partially outside
	    result=1;
	}
	return result;
    }
    // If fully active where instantiations in o
    // have priority
    // returns 0 if it should remain active
    //         1 partially active
    //         -1 inactive
    // parameter (String:DVarHistory)
    int fullyActive(Hashtable view){
	int result=0; //active
	for(int k=0;k<varNr;k++){
	    String crtVariable=variables[k];
	    DVarHistory vHist=(DVarHistory)view.get(crtVariable);
	    if(vHist==null)
		vHist=solver.getExtHist(crtVariable);
	    // not external information
	    if(vHist==null){
		System.out.println("YNogood:fullyActive:not external constraint on "+crtVariable);
		continue;
	    }
	    // completely outside
	    if((vHist.getL()>forbid[k].getU())
	       ||(vHist.getU()<forbid[k].getL()))
		return -1;
	    // internal -no matter if partial or full
	    if(varIDs[k]>=0)
		continue;
	    // inside
	    if((vHist.getL()>=forbid[k].getL())
	       &&(vHist.getU()<=forbid[k].getU()))
		continue;
	    // partially outside
	    result=1;
	}
	return result;
    }
    int active(){
	int result=0; //active
	for(int k=0;k<varNr;k++){
	    // internal
	    if(varIDs[k]>=0){
		Range range=process.getRange(varIDs[k]);
		// completely outside
		if((range.getL()>forbid[k].getU())
		   ||(range.getU()<forbid[k].getL()))
		    return -1;
		continue;
	    }
	    String crtVariable=variables[k];
	    DVarHistory vHist=solver.getExtHist(crtVariable);
	    // not known or interior variable
	    if(vHist==null) continue;
	    // inside
	    if((vHist.getL()>=forbid[k].getL())
	       &&(vHist.getU()<=forbid[k].getU())){
		System.out.println("active:inside");
		continue;
	    }
	    // completely outside
	    if((vHist.getL()>forbid[k].getU())
	       ||(vHist.getU()<forbid[k].getL())){
		System.out.println("active:completely outside");
		return -1;
		}
	    // partially outside
	    result=1;
	}
	System.out.println("active:result="+result);
	return result;
    }
    // activate, deactivate nogood

    // similar staff can be written if accounting
    // of waiting nogoods is wanted
    void activate(){
	for(int k=0; k<varNr; k++){
	    solver.incrNogoods(variables[k],forbid[k]);
	}
    }
    void deactivate(){
	for(int k=0; k<varNr; k++){
	    solver.decrNogoods(variables[k]);
	}
    }
    // if the last setDomain intersects, then
    // setGuilty()
    void setGuiltyIfIntersect(){
	if(intersects) setGuilty();
    }
    void setGuilty(){
	System.out.println("YNogood::setGuilty");
	for(int k=0; k<varNr; k++){
	    solver.setGuilty(variables[k]);
	}
    }
    public void initCover(){
        cover=new ArrayList();
    }
    // called before instantiation
    // sets Guilty
    public void initDomain(Range[] dom){
        initCover();
        System.out.println("Nogood domain:");
	if(dom==null){
	    System.err.println("initDomain:null parameter");
	    return;
	}
        Range[] newDom=(Range[])dom.clone();
        for(int k=0; k<dom.length;k++){
	    if(dom[k]==null){
		System.err.println("initDomain:null parameter dom["+k+"]");
		return;
	    }
            newDom[k]=(Range)dom[k].clone();
        }
        Print(newDom);
	if(dNogood!=null){
	    System.out.println("Nogood info was:");
	    System.out.println(dNogood);
	}
        System.out.println("Forbided domain:");
        Print(forbid);
	if(empty(newDom)){
	    intersects=false;
	    // intersects forbidden
	    System.out.println("YNogood::empty Domain");
	    //setGuilty();
	    return;
	}
        if(intersectRM(newDom,forbid)){
            removeInterior(newDom,forbid);
	    intersects=true;
	    // intersects forbidden
	    System.out.println("YNogood::initDomain");
	    setGuilty();
        } else{
	    intersects=false;
            System.out.println("domain added to nogood");
            cover.add(newDom);
        }
    }
    private void Print(Range[] box){
        for(int k=0; k<box.length; k++){
            if(box[k]!=null)
                System.out.println(variables[k]+"=["+box[k].getL()+"\t,"+box[k].getU()+"\t]");
            else
                System.out.println(variables[k]+"= null");
            System.out.println("\n");
        }
    }
    public int getID(int index){
        return varIDs[index];
    }
    public int getVariablesNr(){
        return varNr;
    }
    public String getVariable(int index){
        return variables[index];
    }
    public boolean getNextBox(Range[] result, Range[] domain){
        Range[] toScan;
	Range tmp;
        while(!cover.isEmpty()){
            toScan=(Range[])cover.remove(0);
	    //perhaps the domain has changed(shrunk)
	    if(intersectRM(toScan,domain)){
		System.out.println("Intersecting candidate");
		Print(toScan);
		for(int k=0;k<result.length;k++){
		    tmp=(Range)toScan[k].clone();
		    tmp.restrict(domain[k]);
		    result[k]=(Range)tmp.clone();
		}
	    }
	    else{
		System.out.println("This do not intersect");
		Print(toScan);
		continue;
	    }
            System.out.println("Found in nogood");
            Print(result);
            return true;
        }
        return false;
    }
    public void setOrder(int[] _score, int[] _order, int[] _invorder){}
    public boolean haveID(int id){
        for(int k=0;k<varNr;k++)
            if(id==varIDs[k])
                return true;
        return false;
    }
    public void setID(int index,int id){
        varIDs[index]=id;
    }
    // called just after initDomain
    public void resetMark(Range[] box){
	if(empty(box)){
	    System.out.println("resetMark:empty box");
	    return;
	}
        ArrayList newCover;
        newCover=cover;
        cover=new ArrayList();
        while(!newCover.isEmpty()){
            Range[] space=(Range[])newCover.remove(0);
            // the space minus the parameter
            // space may be empty!
            if(intersectRM(space,box))
                removeInterior(space,box);
            else
                cover.add(space);
        }
    }
    public void resetContinuationPoints(){
    }
    private boolean empty(Range[] box){
	for(int k=0; k<box.length; k++){
	    if(varIDs[k]<0) continue;
	    if((box[k]==null)||(box[k].size()<=0)){
		if(box[k]==null)
		    System.err.println("YNogood:empty:null=box["+k+"]");
		return true;
	    }
	}
	return false;
    }
    // returns true if the parameters intersect
    private boolean intersectRM(Range[] space1, Range[] space2){
	// at empty domain get next....
	if(empty(space1)||empty(space2)){
	    System.out.println("intersecting empty boxes");
	    return false;
	}
	if(space1.length!=space2.length){
	    System.err.println("Not same length for domain&forbiden");
	    return false;
	}
        boolean result=false;
        for(int k=0;k<space1.length;k++){
            if(varIDs[k]<0) continue;
            // true if at least something
            // common
            result=true;
            if(space1[k].getL()>space2[k].getU())
                return false;
            if(space1[k].getU()<space2[k].getL())
                return false;
        }
	if(!result)
	    System.err.println("intersectRM:Nothing in common!!error");
        return result;
    }
    // the space minus the parameter
    // sent to cover
    // known that they intersect
    private void removeInterior(Range[] space, Range[] box){
        for(int k=0; k<varNr; k++){
            if(varIDs[k]<0) continue;
            if(box[k].getL()>space[k].getL()){
                Range[] dom=(Range[])space.clone();
                split(space,dom,k,box[k].getL()-1);
                System.out.println("Adding domain");
                Print(space);
                cover.add(space);
                space=dom;
            }
            if(box[k].getU()<space[k].getU()){
                Range[] dom=(Range[])space.clone();
                split(space,dom,k,box[k].getU());
                System.out.println("Adding domain");
                Print(dom);
                cover.add(dom);
            }
        }
        System.out.println("Eliminated domain");
        Print(space);   
    }
    // splits a box on direction at value lastOriginal that remains in original
    // The rest is put in "right"
    private void split(Range[] original, Range[] right, int direction, int lastOriginal){
        for(int k=0; k<original.length; k++)
            right[k]=(Range)original[k].clone();
        original[direction].setU(lastOriginal);
        right[direction].setL(lastOriginal+1);
    }
    // meant for original constraints
    public void Print(){}
}
